-- | Parser para el lenguaje imperativo simple. El parser
-- toma como entrada una lista de tokens.
module Parser where

import Lang
import Lexer

import Text.Parsec

type TParser = Parsec [Token] ()

parser :: [Token] -> Either ParseError Comm
parser = runParser pComm () "lexer"

empty :: TParser ()
empty = do s <- getInput
           if null s then return ()
           else fail ""

-- | Un parser muy sencillo para skip
pskip :: TParser Comm
pskip = do _ <- silentKw KSkip
           return Skip

-- | Un parser para if ... then ... else 
pif :: TParser Comm
pif = do _ <- silentKw KIf
         b <- pAssert
         _ <- silentKw Then
         c <- pComm
         _ <- silentKw Else
         c' <- pComm
         _ <- silentKw Fi
         return $ If b c c'

pwhile :: TParser Comm
pwhile = do _ <- silentKw KWhile
            b <- pAssert
            _ <- silentKw Do
            c <- pComm
            _ <- silentKw Od
            return $ While b c

pAssign :: TParser Comm
pAssign = do v <- pvar
             _ <- silentOp TokAssign
             e <- pIntExp
             return $ Assign v e

-- | Parser para comandos que no tienen secuencias. (COMPLETAR)
pComm' :: TParser Comm
pComm' = choice [pif,pskip,pAssign,pwhile]

-- | Parser de secuencias
pseq :: TParser Comm
pseq = do c <- pComm'
          _ <- silentOp TokSemicolon
          c' <- pComm
          return $ Seq c c'

-- | Parser de comandos: o bien una secuencia o bien un comando solito.
pComm :: TParser Comm
pComm = try pseq <|> pComm'


pvar :: TParser Iden
pvar = ptk str
    where str (Str v) = Just v
          str _ = Nothing

-- | Parser para expresiones enteras. (COMPLETAR)
pIntExp :: TParser IntExpr
pIntExp = choice [ var'
                 , num
                 ]
    where var' = do v <- pvar
                    return $ Var v
          num = do n <- ptk isNum
                   return $ Const n
          isNum (Num n) = Just n
          isNum _ = Nothing

-- | Parser para constantes booleanas.
pBoolConst :: TParser Assert
pBoolConst = do b <- pkws [ (KTrue,CTrue) 
                          , (KFalse,CFalse)
                          ]
                return b

-- | Parser para expresiones booleanas. (COMPLETAR)
pAssert :: TParser Assert
pAssert = choice [pBoolConst]

-- | Consumimos un keyword y nada más.
silentKw :: Keyword -> TParser ()
silentKw k = silent isK k
    where isK (Kw k') = Just k'
          isK _ = Nothing

-- | Consumimos un operador y nada más.
silentOp :: Operator -> TParser ()
silentOp o = silent isOp o
    where isOp (Op o') = Just o'
          isOp _ = Nothing

-- | Consumimos un símbolo y nada más.
silentSym :: Symbol -> TParser ()
silentSym s = silent isSym s
    where isSym (Sym s') = Just s'
          isSym _ = Nothing

-- | Si tenemos la suerte que isA para el token actual devuelva justo
-- a', entonces comparamos a con a', en ese caso consumimos el token
-- actual si no, no consumimos nada.
silent :: Eq a => (Token -> Maybe a) -> a -> TParser ()
silent isA a = ptk check
    where check t = case isA t of
                      Just a' -> if a == a' then Just () else Nothing
                      Nothing -> Nothing

-- | 
pkws :: [(Keyword,a)] -> TParser a
pkws cs = ptk l
    where l (Kw k) = lookup k cs
          l _ = Nothing

-- | Dada una función que, con suerte, nos devuelve un a, vemos 
-- un token, y si tuvimos suerte lo consumimos y devolvemos un a.
ptk :: (Token -> Maybe a) -> TParser a
ptk f = tokenPrim show incpos f

-- | Cada vez que consumimos un token incrementamos la posición de
-- nuestro parser (notar que ya perdimos la relación con la posición
-- en el string original).
incpos :: SourcePos -> Token -> [Token] -> SourcePos
incpos p _ _ = incSourceColumn p 1
